/* TECO for Ultrix   Copyright 1986 Matt Fichtenbaum			    */
/* This program and its components belong to GenRad Inc, Concord MA 01742   */
/* They may be copied if this copyright notice is included		    */

/* te_chario.c	 character I/O routines	  02/05/88 */
/* mod to leave stdin modes unchanged if not a terminal */
/* mod 10/21/87 to not reset signals on TTY_SUSP */
/* version for multiple windows 04/13/89 15.54 */
/* mod 7/20/89 to declare signal handlers as "void" not "int" */
/* mod 12/14/90  09.31 for function keys */
/* sun keyboard version  04/10/91  09.36 */
/* sun window version 08/13/91  15.43	 */
/* tecotool version with mouse support  12/02/91  10.56 */

#include <errno.h>
#include <sys/time.h>
#include "te_defs.h"
#include <sunwindow/notify.h>
#include <fcntl.h>
#ifndef DEBUG
#include <signal.h>
extern void int_handler();
extern void stp_handler();
extern void hup_handler();
#define SIGINTMASK 2
#endif

int lf_sw;				    /* nonzero: make up a LF following an entered CR */
int ttyflags;				    /* flags for (stdin) file descriptor */
struct tchars tc_orig, tc_new, tc_noint;    /* original, new, disabled intrpt tty special chars */
struct ltchars lc_orig, lc_new;		    /* original and new local special chars */
struct sgttyb tty_orig, tty_new;	    /* original and new tty flags */
int tty_local;				    /* original tty local mode flags */
int lnoflsh = LNOFLSH;			    /* bit to force "no flush on interrupt */

#ifndef DEBUG
struct sigvec intsigstruc = { int_handler, 0, 0 } ;	/* info structure for ^C interrupt	*/
struct sigvec stpsigstruc = { stp_handler, 0, 0 } ;	/* info structure for "stop" signal	*/
struct sigvec hupsigstruc = { hup_handler, 0, 0 } ;	/* info structure for "hangup" signal	*/
struct sigvec nosigstr = { SIG_DFL, 0, 0 };		/* default structure for signal		*/
#endif

int inp_noterm;		    /* nonzero if standard input is not a terminal */
int out_noterm;		    /* nonzero if standard output is not a terminal */
/* set tty (stdin) mode.  TECO mode is CBREAK, no ECHO, sep CR & LF		    */
/* operation; normal mode is none of the above.  TTY_OFF and TTY_ON do this	*/
/* absolutely; TTY_SUSP and TTY_RESUME use saved signal status.			*/

setup_tty(arg)
    int arg;
{
    extern int errno;
    int ioerr;
    struct sgttyb tmpbuf;
    static void notify_sigvec();
    static Notify_value int_notify();
    static Notify_value stp_notify();
    static Notify_value hup_notify();

/* initial processing: set tty mode */

    if (arg == TTY_ON)
    {
	ioerr = ioctl(fileno(stdin), TIOCGETP, &tty_orig);  /* get std input characteristics */
	inp_noterm = (ioerr && (errno == ENOTTY));	    /* nonzero if input not a terminal */
	ioerr = ioctl(fileno(stdout), TIOCGETP, &tmpbuf);   /* get std output characteristics */
	out_noterm = (ioerr && (errno == ENOTTY));	    /* nonzero if output not a terminal */
	ioctl(fileno(stdout), TIOCLGET, &tty_local);	    /* get current "local mode flags" word */

	ttybuf = tty_new = tty_orig;			    /* make a copy of tty control structure */
	tty_new.sg_flags = (tty_new.sg_flags & ~ECHO & ~CRMOD) | CBREAK;    /* turn on teco modes */

	ioctl(fileno(stdin), TIOCGETC, &tc_orig);	/* read current tchars */
	tc_new = tc_orig;				/* make local copy */
	tc_new.t_quitc = tc_new.t_brkc = -1;		/* disable "quit" and "delimiter" chars */
	tc_noint = tc_new;
	tc_noint.t_intrc = -1;				/* disable the interrupt char in this one */

	ioctl(fileno(stdin), TIOCGLTC, &lc_orig);	/* read current ltchars */
	lc_new = lc_orig;				/* make local copy */
	lc_new.t_rprntc = lc_new.t_werasc = lc_new.t_lnextc = -1;   /* disable "reprint," "word erase," "lit next" */
    }

    if (!inp_noterm)	    /* if std input is from terminal */
    {
	if ((arg == TTY_ON) || (arg == TTY_RESUME))
	{
	    ioctl(fileno(stdin), TIOCSETN, &tty_new);	    /* set flags for teco */
	    ioctl(fileno(stdin), TIOCSETC, &tc_new);	    /* update both */
	    ioctl(fileno(stdin), TIOCSLTC, &lc_new);
	    ioctl(fileno(stdout), TIOCLBIS, &lnoflsh);	    /* disable "interrupt => flush buffers" */

#ifndef DEBUG
	    if (arg == TTY_ON)
	    {
		if (!toolsw)				    /* normal mode */
		{
		    sigvec(SIGTSTP, &stpsigstruc, 0);		/* set up to trap "stop" signal */
		    sigvec(SIGINT, &intsigstruc, 0);		/* and "interrupt" signal */
		    sigvec(SIGHUP, &hupsigstruc, 0);		/* and "hangup" signal */
		}
		else					    /* tecotool mode */
		{
		    notify_sigvec(SIGTSTP, stp_notify);
		    notify_sigvec(SIGINT, int_notify);
		    notify_sigvec(SIGHUP, hup_notify);
		}

	    }
#endif
	}

	else				/* argument is TTY_OFF or TTY_SUSP */
	{
	    ioctl(fileno(stdin), TIOCSETC, &tc_orig);	    /* put both back to orig states */
	    ioctl(fileno(stdin), TIOCSLTC, &lc_orig);
	    ioctl(fileno(stdout), TIOCLSET, &tty_local);    /* restore local mode flags to original states */
	    ioctl(fileno(stdin), TIOCSETN, &tty_orig);	    /* set flags back to original, but don't flush input */
#ifndef DEBUG
	    if (arg == TTY_OFF)
	    {
		if (!toolsw)
		{
		    sigvec(SIGTSTP, &nosigstr, 0);		/* restore default signal handling */
		    sigvec(SIGINT, &nosigstr, 0);
		    sigvec(SIGHUP, &nosigstr, 0);
		}
		else
		{
		    notify_sigvec(SIGTSTP, 0);
		    notify_sigvec(SIGINT, 0);
		    notify_sigvec(SIGHUP, 0);
		}
	    }
#endif
	}
    }
}
/* routines to handle keyboard input */

/* routine to get a character without waiting, used by ^T when ET & 64 is set	*/
/* if lf_sw is nonzero, return the LF; else use the FNDELAY fcntl to inquire of the input */
/* if input is not a terminal don't switch modes */

unsigned short get_kbd_enh();

int gettty_nowait()
{
    int c;

    if (lf_sw)
    {
	lf_sw = 0;
	return(LF);	    /* LF to be sent: return it */
    }
    if (!inp_noterm) fcntl(fileno(stdin), F_SETFL, ttyflags | FNDELAY);	    /* set to "no delay" mode */
    while (!(c = get_kbd_enh()));		/* read character, or -1, skip nulls */
    if (!inp_noterm) fcntl(fileno(stdin), F_SETFL, ttyflags);		    /* reset to normal mode */
    if (c == CR) ++lf_sw;			/* CR: set switch to make up a LF */
    return(c);
}



/* normal routine to get a character */

int in_read = 0;	/* flag for "read busy" (used by interrupt handler) */

int gettty()
{
    int c;

    if (lf_sw)
    {
	lf_sw = 0;
	return(LF);	/* if switch set, make up a line feed */
    }
    ++in_read;				    /* set "read busy" switch */
    while(!(c = get_kbd_enh()));	    /* get character; skip nulls */
    in_read = 0;			    /* clear switch */
    if (c == CR) ++lf_sw;		    /* CR: set switch to make up a LF */
    if (c == EOF) ERROR(E_EFI);		    /* end-of-file from standard input */
    return(c);
}
/* routine to read keyboard characters and interpret incoming escape sequences */

/* sequences accepted :
 * top F1	    $[224z	0412	q`A
 * through				q`BCDEFGHIJK
 * top F12	    $[235z	0425	q`L
 * right R1 (pause) $[208z	0426	q`M
 * right R2 (PrSc)  $[209z	0427	q`N
 * right R3 (break) $[210z	0430	q`O
 * right R4 (=)	    $[211z	0431	q`P
 * right R5 (/)	    $[212z	0432	q`Q
 * right R6 (*)	    $[213z	0433	q`R
 * right R7 (home)  $[214z	0434	q`S
 * right R8 (up)    $[A		0435	q`T
 * right R9 (PgUp)  $[216z	0436	q`U
 * right R10 (left) $[D		0437	q`V
 * right R11 (5)    $[218z	0440	q`W
 * right R12 (rght) $[C		0441	q`X
 * right R13 (end)  $[220z	0442	q`Y
 * right R14 (down) $[B		0443	q`Z
 * right R15 (PgDn) $[222z	0401	q`1
 * right 0   (ins)  $[247z	0402	q`2
 * right .	    $[249z	0403	q`3
 * right ENT	    $[250z	0404	q`4
 * right +	    $[253z	0405	q`5
 * right -	    $[249z	0406	q`6
 *
 * mouse left			0407	q`7
 * mouse middle			0410	q`8
 * mouse right			0411	q`9
 */

struct trans_entry
{
    unsigned char i;
    unsigned char o;
};

#define xx 0xff
#define KEY_MOUSEL 0407
#define KEY_MOUSEM 0410
#define KEY_MOUSER 0411

static unsigned char key_table[] =
{
    xx, xx, xx, xx, xx, xx, xx, xx, 22, 23,	/* 0..9 */
    24, 25, 26, 27, 28, 29, 30, 31, 32, 33,	/* 10..19 */
    34, 35, 1,  xx, 10, 11, 12, 13, 14, 15,	/* 20..29 */
    16, 17, 18, 19, 20, 21, xx, xx, xx, xx,	/* 30..39 */
    xx, xx, xx, xx, xx, xx, xx, 2,  xx, 3,	/* 40..49 */
    4,  xx, xx, 5,  6,  xx, xx, xx, xx, xx	/* 50..59 */
} ;
static unsigned char key_cursor_table[] =
{
    29, 35, 33, 31				/* A..D */
} ;
static struct timeval max_timeout = { 0L, 400000L };	    /* 400 ms */
static unsigned long read_mask = 1;
static unsigned long write_mask = 0;
static unsigned long except_mask = 0;
static unsigned short n_buff;
static unsigned char c2 = 0;
static unsigned char c3 = 0;
static unsigned char c4 = 0;
static unsigned char c5 = 0;
static unsigned char c6 = 0;
static unsigned char mouse_buff[5];
int mouse_data[5];		    /* 0: mode, 1: dot, 2: buffer, 3: Y, 4: x */

unsigned short get_kbd_enh()
{
    unsigned char c1;
    unsigned short retval;
    unsigned short t;
    int conv_hex();
    int xy_to_dot();

    if (inp_noterm) return(getchar());
    else fflush(stdout);
    if (c2 != 0)
    {
	retval = c2;
	c2 = 0;
    }
    else if (c3 != 0)
    {
	retval = c3;
	c3 = 0;
    }
    else if (c4 != 0)
    {
	retval = c4;
	c3 = 0;
    }
    else if (c5 != 0)
    {
	retval = c5;
	c5 = 0;
    }
    else if (c6 != 0)
    {
	retval = c6;
	c6 = 0;
    }
    else
    {
	(void) read(fileno(stdin), &c1, 1);	/* read a byte */
	if ((retval = c1) == 033)			    /* escape */
	{
	    read_mask = 1;
	    c2 = 0;
	    if (select(1, &read_mask, &write_mask, &except_mask, &max_timeout))		 /* char within allowed time */
	    {
		(void) read(fileno(stdin), &c2, 1);	    /* read another */
		if (c2 == '[')				    /* valid second byte */
		{
		    read_mask = 1;
		    c3 = 0;
		    if (select(1, &read_mask, &write_mask, &except_mask, &max_timeout))	    /* timeout again */
		    {
			(void) read(fileno(stdin), &c3, 1);	/* read third char */

			if (c3 >= 'A' && c3 <= 'D')		/* cursor keys */
			{
			    retval = 0400 | key_cursor_table[c3 -'A'];
			    c2 = c3 = 0;
			}
			else if (c3 == '2')			/* other function keys */
			{
			    c4 = 0;
			    read_mask = 1;
			    if (select(1, &read_mask, &write_mask, &except_mask, &max_timeout))
			    {
				(void) read(fileno(stdin), &c4, 1); /* read fourth char */
				if (c4 >= '0' && c4 <= '5')	    /* must be digits */
				{
				    c5 = 0;
				    read_mask = 1;
				    if (select(1, &read_mask, &write_mask, &except_mask, &max_timeout))
				    {
					(void) read(fileno(stdin), &c5, 1);
					if (c5 >= '0' && c5 <= '9')	/* fifth must be digits */
					{
					    c6 = 0;
					    read_mask = 1;
					    if (select(1, &read_mask, &write_mask, &except_mask, &max_timeout))
					    {
						(void) read(fileno(stdin), &c6, 1);
/* tecotool encodes the shift, ctrl, meta */	if (c6 >= ('z' & ~2) && c6 <= ('z' | 5))	/* success */
/* keys in the low 3 bits of the 'z' char */	{
/* this returns them in 100, 200, 400 */	    t = key_table[10 * c4 + c5 - 11 * '0'];
/* of the character */				    if (t != 0xff)	/* valid sequence */
						    {
							retval = t | 0400 | (((c6 & 7) ^ 2) << 9);
							c2 = c3 = c4 = c5 = c6 = 0;
						    }
						}
					    }
					}
				    }
				}
			    }
			}
			else if (c3 == '3')	    /* mouse code */
			{
			    c4 = 0;
			    read_mask = 1;
			    if (select(1, &read_mask, &write_mask, &except_mask, &max_timeout))
			    {
				c2 = c3 = 0;
				(void) read(fileno(stdin), mouse_buff, 5);
				mouse_data[3] = conv_hex(&mouse_buff[0]);	    /* get coordinates */
				mouse_data[4] = conv_hex(&mouse_buff[2]);
				mouse_data[1] = xy_to_dot(mouse_data[4], mouse_data[3], &mouse_data[2]);
				retval = KEY_MOUSEL;
				if ((mouse_buff[4] & 0x10) != 0) retval = KEY_MOUSEM;
				else if ((mouse_buff[4] & 0x20) != 0) retval = KEY_MOUSER;
				retval |= ((mouse_buff[4] & 0x07) << 9 | 0400);
			    }
			}
		    }
		}
	    }
	}
    }
    return(retval);
}


int conv_hex(p)
    char *p;
{
    int retval;

    if (*p >= '0' && *p <= '9') retval = (*p - '0') << 4;
    else if (*p >= 'a' && *p <= 'f') retval = (*p - 'a' + 10) << 4;
    if (*++p >= '0' && *p <= '9') retval |= (*p - '0');
    else if (*p >= 'a' && *p <= 'f') retval |= (*p - 'a' + 10);
    return(retval);
}


/* read and modify mouse parameters */

#define E_IMO E_IFC

void do_fm()
{
    if (!esp->flag1 || esp->val1 < 1 || esp->val1 > 5 ||
	(esp->flag2 && esp->val1 > 1) )
	    ERROR(E_MOU);
    if (esp->flag2)
    {
	mouse_data[0] = esp->val2;
	esp->flag2 = esp->flag1 = 0;
    }
    else esp->val1 = mouse_data[esp->val1 - 1];
    esp->op = OP_START;
}


#ifndef DEBUG

/* routine to handle interrupt signal */

void int_handler()
{

    if (exitflag <= 0)			    /* if executing commands */
    {
	if (et_val & ET_CTRLC) et_val &= ~ET_CTRLC;	/* if "trap ^C" set, clear it and ignore */
	else exitflag = -2;				/* else set flag to stop execution */
    }
    if (in_read)			    /* if interrupt happened in "getchar" pass a ^C to input */
    {
	in_read = 0;					/* clear "read" switch */
	ioctl(fileno(stdin), TIOCSETC, &tc_noint);	/* disable interrupt char */
	qio_char(CTL (C));				/* send a ^C to input stream */
	ioctl(fileno(stdin), TIOCSETC, &tc_new);	/* reenable interrupt char */
    }
}
#endif


/* routine to disable (1), enable (0) ^C interrupt, used to block interrupts during display update */

int old_mask;		    /* storage for previous signal mask */
#define INT_MASK 2

block_inter(func)
    int func;
{
#ifndef DEBUG
    if (func) old_mask = sigblock(INT_MASK);		/* if arg nonzero, block interrupt */
    else sigsetmask(old_mask);				/* otherwise restore old signal mask */
#endif
}



#ifndef DEBUG
/* routine to handle "stop" signal (^Y) */

void stp_handler()
{
    crlf();
    window(WIN_SUSP);		    /* restore screen */
    setup_tty(TTY_SUSP);	    /* put tty back to normal */
    sigvec(SIGTSTP, &nosigstr, 0);  /* put default action back */
    sigsetmask(0);		    /* unblock "suspend" signal */
    kill(0, SIGTSTP);		    /* suspend this process */

/* ----- process gets suspended here ----- */

    sigvec(SIGTSTP, &stpsigstruc, 0);	/* restore local handling of "stop" signal */
    setup_tty(TTY_RESUME);		/* restore tty */
    pbuff->buff_mod = 0;		/* set whole screen modified */
    if (win_data[7])		/* redraw window */
    {
	window(WIN_RESUME);	/* re-enable window */
	window(WIN_REDRAW);	/* force complete redraw */
	window(WIN_REFR);	/* and refresh */
    }
    qio_char('\0');		/* wake up the input */
    if (exitflag) retype_cmdstr('*');	/* if not executing, prompt again and echo command string so far */
}
#endif



/* simulate a character's having been typed on the keyboard */

qio_char(c)
    char c;
{
    ioctl(fileno(stdin), TIOCSTI, &c);		    /* send char to input stream */
}


/* routine to handle "hangup" signal */
#ifndef DEBUG

void hup_handler()
{
    if (!exitflag) exitflag = -3;		    /* if executing, set flag to terminate */
    else
    {
	panic();				    /* dump buffer and close output files */
	exit(1);
    }
}
#endif
/* signal interface to SunWindows notifier */

static Notify_value int_notify()
{
    int_handler();
    return(NOTIFY_DONE);
}

static Notify_value stp_notify()
{
    return(NOTIFY_IGNORED);
}

static Notify_value hup_notify()
{
    hup_handler();
    return(NOTIFY_DONE);
}


static void notify_sigvec(sig, handler)
    int sig;
    void (*handler)();
{
    (void) notify_set_signal_func(sig, handler, sig, NOTIFY_ASYNC);
}


/* type a crlf */

crlf()
{
    type_char(CR);
    type_char(LF);
}



/* reset ^O status */

int lflusho = LFLUSHO;
int lfo;

reset_ctlo()
{
    ioctl(fileno(stdin), TIOCLGET, &lfo);	/* read flags */
    if (lfo & LFLUSHO)				/* if ^O was set */
    {
	ioctl(fileno(stdin), TIOCLBIC, &lflusho);   /* reset ^O */
	crlf();					    /* type a crlf */
    }
}
/* routine to type one character */

type_char(c)
    char c;
{
    void w_linefeed();

    if ((char_count >= WN_width - 1) && (c != CR) && !(spec_chars[c] & A_L))	/* spacing char beyond end of line */
    {
	if (et_val & ET_TRUNC) return;	    /* truncate output to line width */
	else crlf();			    /* otherwise do automatic new line (note recursive call to type_char) */
    }

    if ((c & 0340) == 0)		/* control char? */
    {
	switch (c & 0177)
	{
	    case CR:
		putchar(c);
		char_count = 0;
		break;

	    case LF:
		w_linefeed(1);
		break;

	    case ESC:
		if ((et_val & ET_IMAGE) && !exitflag) putchar(c);
		else
		{
		    putchar('$');
		    char_count++;
		}
		break;

	    case TAB:
		if ((et_val & ET_IMAGE) && !exitflag) putchar(c);
		else for (type_char(' '); (char_count & tabmask) != 0; type_char(' '));
		break;

	    default:
		if ((et_val & ET_IMAGE) && !exitflag) putchar(c);
		else
		{
		    if (char_count >= WN_width - 2) crlf();
		    putchar('^');
		    putchar(c + 'A'-1);
		    char_count += 2;
		}
		break;
	}
    }
    else
    {
	putchar(c);
	char_count++;
    }
}
