/* TECO for Ultrix   Copyright 1986 Matt Fichtenbaum			    */
/* This program and its components belong to GenRad Inc, Concord MA 01742   */
/* They may be copied if this copyright notice is included		    */

/* te_window.c	 window for teco   12/17/87				    */
/* version for multiple buffers  05/31/89 15.17				    */
/* add FL  12/18/90 09.01						    */
/* make FL work with wrapped lines  12/27/90  09.08			    */
/* fix w_setpointer using wrong buffer 01/08/91 10.38			    */
/* version for sun window  08/13/91  15.52				    */
/* version for tecotool with mouse support  12/02/91  09.18		    */

/* This attempts to be a real window, without unecessary redraw */
/* it is very terminal-specific, and ought to be rewritten to be general */

#include "te_defs.h"
void w_linefeed(), w_setline();

/* maximum screen height and width (horiz and vert, not height and vidth) */
#define W_MAX_V 80
#define W_MAX_H 150
#define MAX 0x7fffffff		/* maximum positive integer, for "last modified" location */
#define W_MARK 0200		/* "this loc is special" in screen image */


/* image of current window */

struct w_line		/* data associated with one screen line */
{
    int start, end;		/* dot at beginning, at end */
    short n, cflag, col;	/* number of char positions used, line continuation flag, starting col */
    char ch[W_MAX_H];		/* image of line */
}
     w_image[W_MAX_V];

struct w_line *wlp[W_MAX_V];	/* table of pointers to each line's screen image */

/* data for each buffer window */

struct w_data
{
    short start;		/* first terminal row */
    short end;			/* last terminal row */
    short size;			/* number of rows */
    short half_size;		/* approx. half of window size */
    short last_y;		/* last used row */
    short last_dot;		/* buffer location of cursor */
    short curs_y, curs_x;	/* window cursor position */
    char *curs_p;		/* pointer to cursor loc in window image */
    char curs_c;		/* character that was at cursor */
    short curs_crflag;		/* flag that cursor is on a CR */
} win_1, win_2;		    /* one for each window */

short cmd_win_start;		/* command window start */

/* define "this line is continued" / "this line is a continuation" flags */
#define WF_BEG 1
#define WF_CONT 2
/* define funny font characters */
#ifndef NO_SPEC_FONT
#define WC_DEL 127
#define WC_NL 128
#define WC_LF 129
#define WC_HT 130
#define WC_HTCONT 131
#else
#define WC_DEL 126
#define WC_HT 126
#define WC_NL 126
#define WC_LF 126
#define WC_HTCONT 126
#endif

struct bh *pb0;			/* pointer to buffer - altered by dual-window display */
struct w_data *pw0;		/* pointer to current window data */
struct qp w_p1;			/* pointer for window access to buffer */

short curr_x, curr_y;		/* active character position */
short term_x, term_y;		/* current terminal cursor position */
short redraw_sw;		/* forces absolute redraw */

/* routine to perform simple scope operations */
/* (an attempt to concentrate VT-100 specific things in one place) */

vt(func)
    int func;
{
    short t;
    switch (func)
    {
	case VT_CLEAR:		/* clear screen */
	    fputs("\033[H\033[J", stdout);
	    break;

	case VT_EEOL:		/* erase to end of line */
	    fputs("\033[K", stdout);
	    break;

	case VT_EBOL:		/* erase from beginning of line */
	    fputs("\033[1K", stdout);
	    break;

	case VT_SETSPEC1:	/* reverse video */
	    fputs("\033[7m", stdout);
	    break;

	case VT_SETSPEC2:	/* bright reverse video */
	    fputs("\033[1;7m", stdout);
	    break;

	case VT_SETSPEC3:	/* bright reverse video */
	    fputs("\033[1;4m", stdout);
	    break;

	case VT_CLRSPEC:	/* normal video */
	    fputs("\033[0m", stdout);
	    break;

	case VT_BS1:		/* backspace 1 spot */
	    fputs("\b \b", stdout);
	    break;

	case VT_BS2:		/* backspace 2 spots */
	    fputs("\b \b\b \b", stdout);
	    break;

	case VT_LINEUP:		/* up one line */
	    w_linefeed(-1);
	    break;
    }
}
/* routine to set window parameters */

/* 0: scope type, 1: width, 2: height, 3: seeall, 4: mark position,	*/
/* 5: hold mode, 6: upper left corner position, 7: scroll region size	*/
/* 8: secondary window size, 9: secondary window origin, 10: primary window size (read only) */

/*		   0   1	2	 3  4	  5   6	   7	 8   9  10  */
int win_min[]  = { 4,  20,	4,	 0, 0,	 -1,  1,   0,	 0,  1,  1 } ;  /* min values for window parameters */
int win_max[]  = { 4,  W_MAX_H, W_MAX_V, 1, MAX, 12, -1,  20,	20, -1, -1 } ;  /* max values */
int win_data[] = { 4,  132,	24,	 0, 0,	  0,  0,   0,	 0,  0,  0 } ;  /* window parameters	*/

do_window(ref_flag)
    int ref_flag;		    /* nonzero forces "refresh" operation */
{
    int i;

    if (colonflag && !ref_flag)
    {
	i = get_value(0);   /* get sub-function */
	if ((i < 0) || (i > (sizeof(win_data)/sizeof(int)))) ERROR(E_IWA);
	if (!esp->flag2)    /* it's a "get" */
	{
	    esp->val1 = win_data[i];
	    esp->flag1 = 1;
	}
	else
	{
	    if ((esp->val2 < win_min[i]) || (esp->val2 > win_max[i]))	/* check range */
		ERROR(E_IWA);
	    if (i == 8 && WN_scroll != 0 && esp->val2 != WN_secondary)  /* changing secondary area */
	    {
		WN_secondary = esp->val2;
		window(WIN_INIT);		/* reset window */
	    }
	    else if (i == 2 && WN_scroll != 0)  /* terminal height */
	    {
		    win_max[8] = (WN_height = esp->val2) - WN_scroll - 4;	/* set maximum secondary window */
		    if (win_max[8] < 0) win_max[8] = 0;
		    if (WN_secondary > win_max[8]) WN_secondary = win_max[8];
		    window(WIN_INIT);		/* turn on window */
	    }
	    else if (i == 7)
	    {
		if (esp->val2)
		{
		    WN_scroll = esp->val2;
		    win_max[8] = WN_height - WN_scroll - 4;	/* set maximum secondary window */
		    if (win_max[8] < 0) win_max[8] = 0;
		    if (WN_secondary > win_max[8]) WN_secondary = win_max[8];
		    window(WIN_INIT);		/* turn on window */
		}
		else window(WIN_OFF);		/* turn off window */
	    }
	    win_data[i] = esp->val2;		/* redundant for ~0,7:w, but no harm */
	    esp->flag2 = 0;
	    window(WIN_REDRAW);			/* redraw window */
	}
    }
   else				/* no colon, or ^W command */
    {
	if (esp->flag1 || ref_flag)
	{
	    if (!ref_flag && (esp->val1 == -1000)) redraw_sw = 0;   /* -1000W: "forget that output was done" */
	    else window(WIN_DISP);	/* nW or ^W refreshes window */
	}
	 esp->flag2 = esp->flag1 = 0;	    /* no colon, consume args */
    }
    colonflag = 0;
    esp->op = OP_START;
}


/* routine to update screen size with numbers obtained from environment */
/* (called by main program's initialization)				*/

set_term_par(lines, cols)
    int lines, cols;
{
    if ((lines >= win_min[2]) && (lines <= win_max[2])) win_1.size = win_data[2] = lines, win_1.end = lines - 1;
    if ((cols >= win_min[1]) && (cols <= win_max[1])) win_data[1] = cols;
}


/* window routine.  performs function as indicated by argument			    */
/* WIN_OFF:	disables split-screen scrolling					    */
/* WIN_SUSP:	disables split-screen scrolling temporarily			    */
/* WIN_INIT:	sets up display support if split-screen scrolling enabled, else nop */
/* WIN_RESUME:  re-enables display support					    */
/* WIN_REDRAW:  causes window to be redrawn on next refresh call		    */
/* WIN_REFR:	if scrolling enabled, redoes window, else if ev or es enabled, does */
/*		that, else nop							    */
/* WIN_LINE:	does WIN_REFR unless that wouldn't do anything, in which case	    */
/*		it does effective 1EV output					    */

window(arg)
    int arg;
{
    int i;

    pb0 = pbuff;		    /* start with editor's current buffer */

    switch (arg)
    {
	case WIN_OFF:		    /* final window off */
	case WIN_SUSP:		    /* temp window off */
	    if (WN_scroll)	    /* if reset/clean up */
	    {
		/* full margins, cursor to last line, erase line */
		char_count = 0;
		cmd_win_start = 0;
		w_setline(WN_height-1);
		printf("\033[K");
	    }
	    break;
	case WIN_INIT:		    /* initialize window - find output speed */
	    w_init();		    /* set up screen image buffer */
	    if (WN_scroll) vt(VT_CLEAR);	/* if split-screen is enabled, clear screen */
				    /* (fall through to "resume") */

	case WIN_RESUME:	    /* re-enable window */
	    cmd_win_start = (WN_scroll) ? WN_height - WN_scroll : 0;
	    if (WN_scroll)	    /* set scroll region, cursor to bottom */
	    {
		w_separators();
		w_setline(WN_height-1);
	    }
	    break;

	case WIN_REDRAW:	/* force redraw of window */
	    redraw_sw = 1;
	    break;

	case WIN_LINE:		/* display one line unless window enabled or ev */
	    if (WN_scroll || ev_val) window(WIN_REFR);	    /* if a real window is set, do it */
	    else if (w_setptr(pb0, pb0->dot, &w_p1))		/* set pointer to dot... and if there's a buffer */
	    {
		w_lines(0, &w_p1, &w_p1);	/* get to beginning of line */
		window0(1);			/* and type 1 line */
	    }
	    break;

	case WIN_REFR:		/* if enabled, refresh window; else do ev or es */
	    if (WN_scroll) window1_both();	/* if scrolling enabled, refresh the window */
	    else if ((ev_val) || (es_val && search_flag))   /* else if ev or es, do that */
	    {
		i = (ev_val) ? ev_val : es_val;
		if (w_setptr(pb0, pb0->dot, &w_p1))	/* set a pointer at dot... and if there's a buffer */
		    window0(i - w_lines(1 - i, &w_p1, &w_p1));  /* go back (i-1) lines and ahead (i) lines */
	    }
	    break;

	case WIN_DISP:			/* display buffer independent of whether scroll mode is enabled */
	    window1_both();
	    break;

    }	    /* end of switch */

    fflush(stdout);	    /* send output out */
}		    /* end of window() */
/* routine to type n lines with character at "dot" in reverse video		*/
/* used for ev, es, and <BS> or <LF> as immediate commands		    */
/* starting char position is in w_p1; argument is number of lines	    */

window0(num)
    int num;
{
    int wi;
    char wc;		/* temp char */
    register struct bh *pb = pb0;

    for (wi = w_p1.dot; (num > 0) && (wi < pb->z); wi++)	/* for each character */
    {
	wc = w_p1.p->ch[w_p1.c];	/* get character */

	if ((char_count >= WN_width) && (wc != CR) && !(spec_chars[wc] & A_L))  /* if about to exceed width */
	{
	    if (et_val & ET_TRUNC) goto w0_noprint;	    /* truncate: don't print this */
	    else
	    {
		fputs("\033[K", stdout);    /* <eeol> "DEL space" */
		crlf();
		fputs("\176 ");
		char_count = 2;
		--num;			/* one fewer line remaining */
	    }
	}

	if (wi == pb->dot)	    /* if this char is at the pointer */
	{
	    vt(VT_SETSPEC2);		/* set reverse video */
	    if (wc == TAB)
	    {
		type_char(' ');	    /* illuminate the first sp of a tab */
		vt(VT_CLRSPEC);	    /* clear reverse video */
		if (char_count & tabmask) type_char(TAB);
	    }
	    else			/* not a tab */
	    {
		if ((wc == CR) && (char_count < WN_width))  /* CR at rh margin: don't display cursor */
		{
		    type_char(' ');	/* cr: put a space after line */
		    vt(VT_EEOL);	/* erase to eol */
		}
		type_char(wc);		/* type the char, or exec CR */
		if (wc == LF)
		{
		    putchar(WC_LF);
		}
		vt(VT_CLRSPEC);	    /* clear reverse video */
	    }
	}
	else			/* this is not char at pointer */
	{
	    if (wc == CR && curr_x < WN_width) vt(VT_EEOL);	/* erase to EOL */
	    type_char(wc);
	}
	if ((wc == FF) || (wc == VT))	    /* FF & VT end a line */
	{
	    vt(VT_EEOL);	    /* erase rest of this line */
	    crlf();		    /* and leave a blank one */
	    if (!(ez_val & EZ_NOVTFF)) --num;	    /* if FF and VT count as line sep's, count them */
	}

      w0_noprint:
	if (++w_p1.c > CELLSIZE-1) w_p1.p = w_p1.p->f, w_p1.c = 0;  /* next char */
	if (wc == LF) --num;	/* if this is a line feed, count lines */
    }

    if (pb->dot == pb->z) fputs("\033[1;7m \033[0m\033[0K", stdout);	/* type one space and erase rest of line */
    else fputs("\033[0K", stdout);	    /* else just erase to EOL */
}



/* routine to draw two windows */

struct bh *old_pbuff;

window1_both()
{
    pw0 = &win_1;
    pb0 = &buffs[0];
    window1();			/* display primary window */

    if (WN_secondary)
    {
	pw0 = &win_2;
	pb0 = &buffs[1];
	window1();		/* display secondary window */
    }

    if (WN_scroll)		/* cleanup */
    {
	printf("\033[%d;%dH", WN_height, char_count+1);		    /* cursor to bottom */
    }
    else printf("\033[H");	/* no split screen: set home */

    fflush(stdout);		/* flush output */
    WN_origin = wlp[win_1.start]->start;			    /* save first char pos on screen */
    if (WN_secondary) WN_sec_origin = wlp[win_2.start]->start;	    /* in each window */
    redraw_sw = 0;		/* mark screen as updated */
    old_pbuff = pbuff;		/* save old pbuff */
}

/* routine to maintain the screen window					*/
/* if scroll mode is enabled, the VT100 screen is split and only the upper part */
/* is used by this routine; else the whole screen is used.			*/

window1()
{
    int i, j, m, lflag;
    register struct bh *pb = pb0;
    register struct w_data *pw = pw0;

    /* return if nothing has changed */
    if (!redraw_sw && (pbuff == old_pbuff) && (pb->dot == pw->last_dot) && (pb->buff_mod == MAX)) return;

    block_inter(1);			    /* disable ^C interrupts */
    printf("\033[%d;0H", pw->start+1);	    /* "home" for this window */
    curr_y = term_y = pw->start;	    /* indicate cursor is at home */
    term_x = 0;

/* check forced redraw, or z before start of screen */

    if ((redraw_sw) || (pb->z <= wlp[pw->start]->start) || (pw->size <= 1)) window1_abs();

/* check whether pointer is before modified buffer location */

    else if (pb->buff_mod >= pb->dot)	/* yes */
    {

	if (pb->dot < wlp[pw->start]->start)		/* if dot is before screen */
	{
	    w_setptr(pb, wlp[pw->start]->start, &w_p1);	    /* get to beginning of screen */

	/* check whether screen begins with the last part of a continued line */
	    for (j = 0; (wlp[pw->start+j]->cflag & WF_CONT) && (j < pw->half_size); j++);
	    if (j < pw->half_size)	    /* if so, does it continue less than halfway down the screen? */
	    {

		if (j)			    /* is there a partial line? */
		{
		    w_lines(0, &w_p1, &w_p1);	    /* 0L */
		    j -= w_lines(1, &w_p1, NULL);	/* now j is number of display lines before screen */
		}
	/* now look for how many lines back "dot" is: if screen starts with partial line, w_p1 has already been moved */
	/* to beginning of the line and j equals the count of extra lines to scroll */

		for (i = 0; (pb->dot < w_p1.dot) && (i < pw->half_size); ) i -= w_lines(-1, &w_p1, &w_p1);
		if ((pb->dot >= w_p1.dot) && (i < pw->size))		/* found point within reason */
		{
		    w_scroll(j - i);		/* scroll screen down that many lines */
		    wlp[ curr_y = pw->start ]->cflag = wlp[pw->start]->col = curr_x = 0;    /* start from top of window */
		    wlp[pw->start]->start = w_p1.dot;	/* save starting char position */
		    window2(0);			/* and rewrite screen */
		}

		else window1_abs();		/* farther back than that - redraw */
	    }

	    else window1_abs();			/* continuation was too long: give up and redraw */
	}		/* end of "dot is before screen" */

	else if (pb->dot <= wlp[pw->last_y]->end) window1_inc(pb->dot);	    /* on screen - redraw incrementally */

	else window1_after();		    /* dot is after screen: scroll or redraw */
    }		    /* end of "dot is before modified point" */

/* the modified point in the buffer is before dot */
    else
    {
	if (pb->buff_mod < wlp[pw->start]->start) window1_abs();    /* modified point before screen - redraw fully */

	else if (pb->buff_mod <= wlp[pw->last_y]->end)		/* modified point on screen */
	{
	    for (m = 0; pb->buff_mod > wlp[pw->start+m]->end; m++); /* find line with buff_mod */
	    w_setptr(pb, wlp[pw->start+m]->start, &w_p1);		/* set a pointer to start of line with buff_mod */
	    j = (m < pw->half_size) ? pw->size - 1 - m : pw->half_size; /* maximum # of lines between buff_mod & dot */
	    for (i = 0; (pb->dot >= w_p1.dot) && (w_p1.dot < pb->z) && (i <= j); )
		i += (lflag = w_lines(1, &w_p1, &w_p1) ) ? lflag : 1;	/* count lines from buff_mod to first line after dot */
	    if (i > j) window1_abs();			/* too far - redraw */
	    else
	    {
		if (lflag && (pb->dot == pb->z)) i++;		/* if at end, following a LF */
		w_setptr(pb, wlp[pw->start+m]->start, &w_p1);	/* pointer to start of area to redraw */
		if (i > pw->size - m)			/* if there are not enough blank lines on screen */
		    w_scroll(i = i - pw->size + m), curr_y = pw->start + m - i, pw->curs_y -= i;    /* scroll up the difference */
		else curr_y = pw->start + m;
		curr_x = (wlp[curr_y]->cflag & WF_CONT) ? 2 : wlp[curr_y]->col;	    /* line starts at left unless contin. */
		if ((curr_y > pw->curs_y) && (pw->curs_y >= 0)) w_rmcurs();	/* remove old cursor if it won't be written over */
		window2(0);			    /* rewrite newly cleared region */
		for (curr_x = 0; ++curr_y <= pw->end; )	    /* clear rest of screen if needed */
		{
		    wlp[curr_y]->cflag = 0;
		    if (wlp[curr_y]->n) wlp[curr_y]->n = 0, vtm(VT_EEOL);
		}
	    }
	}	    /* end "modified point on screen */

	else window1_after();	    /* modified point after screen: scroll or redraw as appropriate */
    }
    pb->buff_mod = MAX;		    /* done */
    pw->last_dot = pb->dot;
    block_inter(0);		/* reenable interrupts */
}
/* routine to redraw screen absolutely */

window1_abs()
{
    int i, j;
    struct bh *save_pbuff;				/* save old pbuff during set_pointer call */
    register struct bh *pb = pb0;
    register struct w_data *pw = pw0;

    wlp[ curr_y = pw->start ]->col = curr_x = 0;    /* y to start of window, x to 0 */
    save_pbuff = pbuff;					/* we need to use set_pointer in case there's no buffer */
    pbuff = pb;						/* but it needs to point to our buffer */
    set_pointer(pb->dot, &w_p1);			/* make a text buffer, if none, and refresh the display */
    pbuff = save_pbuff;
    w_lines(0, &w_p1, &w_p1);				/* do 0L */
    if ((i = w_lines(pw->half_size, &w_p1, NULL)) == 0) i = 1;	    /* check how many lines after dot */
    if (i > pw->half_size) i = pw->half_size;		/* limit amount after dot */
    for (j = 0; (j < pw->size - i) && (w_p1.dot > 0); )	    /* find start of display area */
	j -= w_lines(-1, &w_p1, &w_p1);
    if (j > pw->size - i) w_lines(1, &w_p1, &w_p1);	    /* if too far back, move up one line */

    wlp[pw->start]->start = w_p1.dot;	    /* indicate where first window line starts */
    window2(0);			    /* refresh the whole display */

    for (curr_x = 0; ++curr_y <= pw->end; )	/* blank out lines not written by window2 */
	if (wlp[curr_y]->n || redraw_sw) wlp[curr_y]->n = 0, vtm(VT_EEOL);
}




/* redraw screen incrementally */

window1_inc(wd)
    int wd;			/* argument is earliest change */
{
    short temp_y;
    register struct bh *pb = pb0;
    register struct w_data *pw = pw0;

/* find the line containing the character at wd */

    for (temp_y = pw->start; wd > wlp[temp_y]->end; temp_y++);

    if ((pw->curs_y != temp_y) || (pb->buff_mod == MAX) || pw->curs_crflag)	/* if the cursor line won't be rewritten */
	w_rmcurs();		    /* remove the old cursor */
    curr_y = temp_y;		    /* and go to work on the beginning of the line with dot */
    curr_x = (wlp[curr_y]->cflag & WF_CONT) ? 2 : wlp[curr_y]->col;	/* line starts at left unless continuation */

    w_setptr(pb, wlp[curr_y]->start, &w_p1);	    /* make a pointer there */
    window2(pb->buff_mod == MAX);	    /* if buffer not modified, redraw only the line with dot */

    if (pb->buff_mod < MAX)		    /* if buffer has changed, erase display lines beyond end of buffer */
	for (curr_x = 0; ++curr_y <= pw->end; )
	    if ( ((wlp[curr_y]->start >= pb->z) || (wlp[curr_y]->start <= wlp[curr_y-1]->end)) &&
		(wlp[curr_y]->n || redraw_sw) )
		    wlp[curr_y]->n = 0, vtm(VT_EEOL), wlp[curr_y]->cflag = 0;
}
/* routine to move window downwards: scroll up or redraw as appropriate */

window1_after()
{
    int i, lflag;
    register struct bh *pb = pb0;
    register struct w_data *pw = pw0;

    w_rmcurs();			    /* remove old cursor */
    w_setptr(pb, wlp[pw->end]->start, &w_p1);	    /* set pointer to start of last line on screen */

    for (i = 0; (pb->dot >= w_p1.dot) && (w_p1.dot < pb->z) && (i <= pw->half_size); )
	i += (lflag = w_lines(1, &w_p1, &w_p1)) ? lflag : 1;	/* fwd one line at a time until > dot or end of buffer */

    if (i <= pw->half_size)	    /* found within n lines */
    {
	if (lflag && (pb->dot == pb->z)) ++i;		    /* if dot is at end of buffer after a LF */
	if (i >= pw->end - pw->last_y)		    /* if there are not enough blank lines on screen */
	    w_scroll(i - (pw->end+1) + pw->last_y), curr_y = pw->end+1 - i; /* scroll up the difference */
	else curr_y = pw->last_y;

	while (curr_y && (wlp[curr_y]->cflag & WF_CONT)) --curr_y;	/* get to start of cont'd lines */
	w_setptr(pb, wlp[curr_y]->start, &w_p1);	/* pointer to start of area to redraw */
	curr_x = wlp[curr_y]->col;		    /* redraw starts at line's first column */
	window2(0);				    /* rewrite newly cleared region */
    }

    else window1_abs();			    /* move down is too far: redraw fully */
}



/* routine to remove the existing cursor */

w_rmcurs()
{
    register struct w_data *pw = pw0;

    if (pw->curs_c)	    /* if there was a cursor */
    {
	w_move(pw->curs_y, pw->curs_x);			/* go remove the old cursor */
	if (pw->curs_c & W_MARK) putchar(pw->curs_c);	    /* if prev char was a spec char */
	else putchar(*pw->curs_p = pw->curs_c);		/* put back the char that was there */
	++term_x;					/* and keep the terminal cursor loc. happy */
    }
}
/* routine to do actual display refresh						*/
/* called with w_p1 at starting char, curr_y, curr_x at starting coordinate	    */
/* rewrites to end of screen if arg = 0, or only until line with cursor if arg = 1  */

window2(arg)
    int arg;
{
    register struct bh *pb = pb0;
    register struct w_data *pw = pw0;
    register int wdot;
    register char wc;		/* temp char */
    register short dflag;	/* nonzero if this is char at dot */
    short cr_found;		/* indicates a cr found on this line */

    cr_found = 0;		/* clear "cr" flag in first line written */
    for (wdot = w_p1.dot; (curr_y <= pw->end) && (wdot < pb->z); wdot++)	/* for each character */
    {
	wc = w_p1.p->ch[w_p1.c] & 0177;	    /* get character */
	if (dflag = (wdot == pb->dot)) if (arg) arg = -1;	/* save "this is char at dot", "on line with dot" */

	if (wc < ' ') switch (wc)		/* dispatch control characters */
	{
	    case CR:
		if (dflag)	    /* if cursor on this CR */
		{
		    if (curr_x < WN_width) w_makecurs(' ', 1), w_type(' ', 1);  /* display a space, unless at end */
		    else pw->curs_crflag = pw->curs_c = 0;		/* else set "no cursor displayed" */
		}
		/* trim remainder of line if this is first cr and old line was longer */
		if (!cr_found && ((curr_x < wlp[curr_y]->n) || redraw_sw))
		{
		    wlp[curr_y]->n = curr_x;
		    if (curr_x < WN_width) vtm(VT_EEOL);
		}
		cr_found = 1;		/* set cr flag */
		wlp[curr_y]->cflag &= ~WF_BEG;	    /* this line is not continued */
		while (curr_y && (wlp[curr_y]->cflag & WF_CONT)) --curr_y;	/* if line is a continuation, scan up */
		curr_x = 0;
		break;

	    case TAB:
		if (curr_x >= WN_width)
		{
		    if (et_val & ET_TRUNC) goto noprint;
		    if (w_overflow(wdot)) goto w2_exit;		    /* extend line */
		}
		if (dflag) w_makecurs( (WN_seeall) ? ('b' | W_MARK) : ' ', 0);
		if (WN_seeall)
		    w_type(WC_HT, 1);				    /* type "HT" char */
		else w_type(' ', dflag);			    /* type one space */
		if (dflag)
		{
		    vt(VT_CLRSPEC);	/* end reverse video */
		    dflag = 0;
		}
		while ((curr_x & tabmask) && (curr_x < WN_width))   /* finish tab */
		     w_type( (WN_seeall) ? (WC_HTCONT | W_MARK) : ' ', 0);
		break;
	    case LF:
		while ((curr_y <= pw->end) && (wlp[curr_y]->cflag & WF_BEG)) ++curr_y;  /* last screen row of this line */
		wlp[curr_y]->end = wdot;	/* save char position that ended this line */
		if (dflag)	/* if this LF is at dot */
		{	    /* put cursor there, save char that was there */
		    w_makecurs( (curr_x < wlp[curr_y]->n) ? wlp[curr_y]->ch[curr_x] : ' ', 0);
		    w_type(WC_LF, 1);		    /* put in a "LF" char */
		}	    /* if no cr found and not in last column, erase rest of line */
		if (!cr_found && (curr_x < wlp[curr_y]->n))
		{
		    wlp[curr_y]->n = curr_x;
		    if (curr_x < WN_width) vtm(VT_EEOL);
		}
		if (dflag) --curr_x;		/* put the cursor back before the artificial LF char, if any */
		if (curr_y >= pw->end)		/* if at end of screen, exit, but... */
		{
		    if (dflag) vt(VT_CLRSPEC);  /* if cursor is here, clear reverse video first */
		    goto w2_exit;
		}

		if ((wlp[curr_y]->cflag & WF_CONT) && (wlp[curr_y]->end - wlp[curr_y]->start == 1))
				    /* if a now-empty cont. line, */
		{							    /* flush it */
		    if (curr_y > 0) wlp[curr_y-1]->cflag &= ~WF_BEG;	    /* remove "cont'd" flag from prev line */
		    arg = 0;							/* and force redraw of rest of screen */
		    if (pw->curs_y == curr_y) pw->curs_c = 0;			/* if cursor was on this line, it will disappear */
		}
		else ++curr_y;		    /* down one line if not absorbing blank contin. line */

		wlp[curr_y]->start = wdot + 1;		    /* assume line starts with next char */
		wlp[curr_y]->col = curr_x;		    /* save starting column */
		cr_found = 0;
		wlp[curr_y]->cflag &= ~WF_CONT;		    /* next line is not a continuation */
		if (curr_x) w_ebol();			    /* if not at left margin, erase beginning of line */
		if (arg == -1)				    /* finished line with dot... quit if spec'd */
		{
		    if (dflag)			    /* but first, if at cursor, clear reverse video */
		    {
			vt(VT_CLRSPEC);
			dflag = 0;
		    }
		    return;
		}
		break;

	    case ESC:
		if (curr_x >= WN_width)
		{
		    if (et_val & ET_TRUNC) goto noprint;
		    if (w_overflow(wdot)) goto w2_exit;		    /* extend line */
		}
		if (dflag) w_makecurs('$', 0);
		w_type('$', dflag);
		break;
	    default:			/* all other control chars print as ^X */
		if (curr_x >= WN_width - 1)
		{
		    if (et_val & ET_TRUNC) goto noprint;
		    if (w_overflow(wdot)) goto w2_exit;
		}
		if (dflag) w_makecurs('^', 0);
		w_type('^', dflag);		/* ^ */
		if (dflag)
		{
		    vt(VT_CLRSPEC);	/* if at cursor, clear reverse video */
		    dflag = 0;
		}
		w_type(wc | 0100, 0);
		break;
	}		    /* end "switch" */
	else			/* a printing character */
	{
	    if (curr_x >= WN_width)
	    {
		if (et_val & ET_TRUNC) goto noprint;
		if (w_overflow(wdot)) goto w2_exit;		/* extend line */
	    }
	    if (dflag) w_makecurs(wc, 0);
	    w_type(wc, dflag);
	}

	if (dflag)
	{
	    vt(VT_CLRSPEC);		/* if at cursor, clear reverse video */
	}

	if ((wc == FF) || (wc == VT))		/* these chars leave a display line blank */
	{
	    if (redraw_sw || (curr_x < wlp[curr_y]->n))
	    {
		wlp[curr_y]->n = curr_x;
		if (curr_x < WN_width) vtm(VT_EEOL);	/* erase rest of line */
	    }
	    wlp[curr_y]->end = wdot;
	    if (curr_y >= pw->end) goto w2_exit;	/* quit if overflow screen */
	    wlp[++curr_y]->start = wdot + 1;
	    cr_found = wlp[curr_y]->cflag = 0;	    /* init new line */
	    if (curr_x -= 2) w_ebol();	    /* back up over ^X; if not at left margin, erase beginning of line */
	    wlp[curr_y]->col = curr_x;		    /* save starting column */
	}
     noprint:
	if (++ w_p1.c > CELLSIZE - 1) w_p1.p = w_p1.p->f, w_p1.c = 0;	/* next char in buffer */
    }	    /* end of "for all characters" */

    if (pb->dot == pb->z)
    {
	if (curr_x < WN_width) w_makecurs(' ', 1), w_type(' ', 1), vt(VT_CLRSPEC);  /* display a space, unless at end */
	else pw->curs_crflag = pw->curs_c = 0;		    /* else set "no cursor displayed" */
    }

    /* clear rest of line if needed */
    if (!cr_found && (redraw_sw || (curr_x < wlp[curr_y]->n)))
    {
	wlp[curr_y]->n = curr_x;
	if (curr_x < WN_width) vtm(VT_EEOL);
    }
    wlp[curr_y]->end = wdot;	    /* save char at end of last line */
  w2_exit:
    pw->last_y = curr_y;		/* record last used line on screen */
}
/* routine to move cursor to current location and then call vt */

vtm(arg)
    int arg;
{
    w_move(curr_y, curr_x);
    vt(arg);
}




/* routine to set reverse video and save cursor location */
/* first argument is char at cursor, 2nd is value for curs_crflag */

w_makecurs(wc, crflag)
    char wc;
    short crflag;
{
    register struct w_data *pw = pw0;

    pw->curs_y = curr_y, pw->curs_x = curr_x, pw->curs_c = wc;  /* save cursor coord and char */
    pw->curs_p = &wlp[curr_y]->ch[curr_x];		/* save location of cursor spot in window image */
    pw->curs_crflag = crflag;	    /* save crflag */
    vt( pb0 == pbuff ? VT_SETSPEC2 : VT_SETSPEC3);	    /* set flag and cursor (bright if active buffer) */
}




/* routine to handle line overflow */
/* returns nonzero if at end of screen, zero otherwise */
/* arg is current character position */

int w_overflow(wd)
{
    register struct w_data *pw = pw0;

    wlp[curr_y]->end = wd-1;		/* last character was end of this line */
    if (wlp[curr_y]->n > curr_x)
    {
	wlp[curr_y]->n = curr_x;
	if (curr_x < WN_width) vtm(VT_EEOL);	    /* if old line was wider, erase */
    }
    if (curr_y >= pw->end) return(1);
    wlp[curr_y]->cflag |= WF_BEG;		/* mark this line as "continued" */
    wlp[++curr_y]->cflag = WF_CONT;		/* next line is a continuation line */
    wlp[curr_y]->start = wd;			/* char about to be printed is this line's first */
    wlp[curr_y]->col = curr_x = 0;		/* new line starts at left margin */
    w_type(WC_NL, W_MARK);			    /* "NL" space */
    w_type(' ', W_MARK);
    return(0);
}
/* routine to type one character:  arguments are char and a */
/* "mark" bit.  If mark is set, the char is always retyped  */

w_type(c, m)
    char c;
    int m;
{
    register char *p;
    register struct w_data *pw = pw0;

    p = &wlp[curr_y]->ch[curr_x];	/* pointer to char image */
    if ((c != *p) || (m) || (redraw_sw) || (curr_x >= wlp[curr_y]->n))
    {
	w_move(curr_y, curr_x);
	putchar(c);
	*p = (m) ? c | W_MARK : c;
	++term_x;
    }
    ++curr_x;
    if (wlp[curr_y]->n < curr_x) wlp[curr_y]->n = curr_x;   /* if we've lengthened the line, record that fact */
}




/* initialize display image */

w_init()
{
    short i, j;

    win_2.start = WN_primary = win_1.size = WN_height - WN_scroll - WN_secondary;	/* set window definition parameters */
    win_1.start = 0;
    win_1.end = win_1.size - 1;
    win_2.size = WN_secondary;
    win_2.end = win_1.end + win_2.size;

/* do borders */

    if (WN_scroll > 0)
    {
	--win_1.size, --win_1.end;	    /* trim for border at command window */
	--win_2.start, --win_2.end;	    /* and move sec window up 1 */
	if (WN_secondary > 0)		    /* if two windows */
	    --win_1.size, --win_1.end;	    /* take middle border from top window */
    }
    win_1.half_size = (win_1.size + 1)/2;
    win_2.half_size = (win_2.size + 1)/2;

    for (i = win_1.start; i <= win_1.end; i++)  /* for each row in primary window */
    {
	wlp[i] = &w_image[i];			/* set pointer to this line's data */
	w_image[i].n = w_image[i].cflag = 0;	/* no chars used, cr flag clear */
	for (j = 0; j < W_MAX_H; w_image[i].ch[j++] = ' ');	/* clear line */
    }

    for (i = win_2.start; i <= win_2.end; i++)  /* for each row in secondary window */
    {
	wlp[i] = &w_image[i];			/* set pointer to this line's data */
	w_image[i].n = w_image[i].cflag = 0;	/* no chars used, cr flag clear */
	for (j = 0; j < W_MAX_H; w_image[i].ch[j++] = ' ');	/* clear line */
    }
}


/* write separator lines between windows on screen */

w_separators()
{
    int i;

    if (WN_scroll > 0)
    {
	printf("\033[%d;0H", win_2.end + 2);	    /* cursor to first line after window 1 */
	for (i = 0; i < WN_width; i++) putchar('=');	/* write a centered line */
    }
    if (WN_secondary > 0)
    {
	printf("\033[%d;0H", win_1.end + 2);	    /* cursor to first line after window 2 */
	for (i = 0; i < WN_width; i++) putchar('-');	/* write a centered line */
    }
}


/* move terminal cursor to stated y, x position */

w_move(y, x)
    short y, x;
{
    printf("\033[%d;%dH", (term_y = y) + 1, (term_x = x) + 1);	    /* absolute jump */
}
/* scroll screen: argument is count: + up, - down */

w_scroll(count)
    int count;
{
    register int i, ic;
    struct w_line *p[W_MAX_V];			/* temp copy of pointer array */
    register struct w_data *pw = pw0;

    if (count > 0)	/* scrolling up */
    {
	for (i = 0; i < count; i++)
	    wlp[pw->start + i]->n = 0;	    /* scroll terminal, blank out image line */
	w_move(pw->start, 0);			/* cursor to top of window */
	printf("\033[%dM", count);		/* delete N lines */
	w_move(pw->end + 1 - count, 0);		/* cursor to where first blank line should be */
	printf("\33[%dL", count);		/* insert N lines */
    }
    else	/* scroll down */
    {
	for (i = 0; i > count; i--)
	    wlp[pw->end+i]->n = 0;
	w_move(pw->end + 1 + count, 0);		/* cursor to first line deleted */
	printf("\033[%dM", -count);		/* delete N lines */
	w_move(pw->start, 0);			/* cursor to top of window */
	printf("\033[%dL", -count);		/* insert N lines */
    }
    for (i = 0; i < pw->size; i++) p[i] = wlp[pw->start + ((pw->size + i + count) % pw->size)]; /* rearrange */
    for (i = 0; i < pw->size; i++) wlp[pw->start + i] = p[i];
}



/* clear line to left of curr_x */
/* if some chars nonblank, does erase from start of line */

w_ebol()
{
    short i, j;
    register struct w_data *pw = pw0;

    for (j = i = 0; i < curr_x; i++) if (wlp[curr_y]->ch[i] != ' ') wlp[curr_y]->ch[i] = ' ', j++;
    if (j || redraw_sw) w_move(curr_y, curr_x-1), vt(VT_EBOL);
}

/* routines to simulate command window scrolling	    */
/* w_setline(n) leaves cursor at line n on absolute screen  */
/* w_linefeed(n) moves n lines in command window (+ or -)   */

static int w_curline = 0;	/* current line */

void w_setline(n)
    int n;			/* number of current line */
{
    if ((w_curline = n) < 0) w_curline = 0;
    else if (w_curline >= WN_height) w_curline = WN_height-1;
    printf("\033[%d;%dH", w_curline+1, char_count+1);
}

void w_linefeed(n)
    int n;
{
    if (n > 0)
    {
	if ((w_curline + n) < WN_height)	/* no scroll needed */
	{
	    printf("\033[%dB", n);		/* down n lines */
	    w_curline += n;
	}
	else
	{
	    n -= (WN_height - w_curline - 1);	/* n is number to scroll */
	    printf("\033[%d;0H\033[%dM\033[%d;%dH",
		cmd_win_start+1,		/* to top of command window */
		n,				/* delete n lines */
		(w_curline = WN_height-n)+1,	/* to start of cleared area */
		char_count+1 );			/* at current column */
	}
    }
    else
    {
	if ((w_curline + n) < cmd_win_start)	/* if moving up too far */
	    n = cmd_win_start - w_curline;
	printf("\033[%dA", -n);			/* move up n lines */
	w_curline += n;
    }
}
/* routine to set a pointer to a given location (like set_pointer) */
/* returns nonzero if a text buffer exists, otherwise 0 */

int w_setptr(pbuff, loc, pp)
    struct bh *pbuff;		    /* current buffer header */
    register int loc;		    /* location */
    register struct qp *pp;	    /* address of pointer */
{
    register int i;

    if (pbuff->f)
    {
	for (i = loc / CELLSIZE, pp->p = pbuff->f; i > 0; i--) pp->p = pp->p->f;
	pp->c = loc % CELLSIZE;
	pp->dot = loc;
    }
    return( (int) pbuff->f);
}
/* routine to move N lines (back, forward, or 0)		*/
/* w_lines(n, &source, &dest) where n is the line count, source */
/* points to a qp at the current pointer, dest, if nonzero,	*/
/* it points to a qp where the result is to go.			*/
/* routine returns actual number of display lines		*/

struct qp w_lines_p;		    /* to compute # of display lines in -N lines */

int w_lines(n, ps, pd)
    int n;			    /* number of lines */
    register struct qp *ps, *pd;    /* source, destination qp's */
{
    register struct buffcell *tp;   /* local copy of the qp */
    register int tc, tdot, tn;
    int tcnt, tl;		    /* chars/line and display line count */
    char tch;

    tdot = ps->dot;
    tp = ps->p;
    tc = ps->c;

    if (n > 0)		/* argument is positive */
    {
	for (tcnt = tl = tn = 0; (tn < n) && (tdot < pb0->z); tdot++)	/* forward over N line separators */
	{
	    if (spec_chars[ tch = tp->ch[tc] ] & A_L) ++tl, ++tn;	/* count separators */
	    else if (!(et_val & ET_TRUNC))	/* if text lines can overflow screen lines */
	    {
		if (!(tch & 0140))		/* if character is a control char */
		{
		    if (tch == CR)					/* CR resets count */
		    {
			if (tcnt > WN_width) ++tl;
			tcnt = 0;
		    }
		    else if (tch == TAB) tcnt = (tcnt | tabmask) +1;	/* tab to next tab stop */
		    else if (tch == ESC) ++tcnt;			/* ESC takes one space */
		    else tcnt += 2;					/* normal control chars take 2 spaces */
		}
		else ++tcnt;		    /* not a control char: takes one space */
		if (tcnt > WN_width) ++tl, tcnt = 2;	    /* if overflow, one more line */
	    }

	    if (++tc > CELLSIZE-1) tp = tp->f, tc = 0;	    /* next character position */
	}
	if (tl > tn) tn = tl;	    /* if counting display lines and there's more of them, return that */
    }
   else		/* argument is zero or negative */
    {
	for (tn = 0; (tn >= n) && (tdot > 0); )	    /* back up over (n+1) line feeds */
	{
	    --tdot;
	    if (--tc < 0) tp = tp->b, tc = CELLSIZE -1;
	    if (spec_chars[tp->ch[tc]] & A_L) --tn;
	}
	if (tn < n)	    /* if stopped on a line sep, fwd over it */
	{
	    ++tn;
	    ++tdot;
	    if (++tc > CELLSIZE-1) tp = tp->f, tc = 0;
	}

	if (!(et_val & ET_TRUNC) && (n != 0))		/* if text line can overflow display line */
	{
	    w_lines_p.dot = tdot, w_lines_p.p = tp, w_lines_p.c = tc;	    /* then count the number of display */
	    tn = -w_lines(-n, &w_lines_p, 0);		    /* lines in the N text lines we just backed up over */
	}
    }

    if (pd) pd->dot = tdot, pd->p = tp, pd->c = tc;	/* if an "after" pointer given, update it */
    return(tn);
}
/* routine to move n chars on line */
/* call with # of columns, return new dot value */
int x_to_dot(x)
    int x;
{
    struct qp m_dot;
    int m_count;
    int m_char;

    (void) w_setptr(pbuff, pbuff->dot, &m_dot);	    /* set pointer to first char on line */
    for (m_count = 0; m_count < x && m_dot.dot < pbuff->z; fwdc(&m_dot))    /* for each character on line */
    {
	m_char = m_dot.p->ch[m_dot.c];		/* character */
	if (m_char < ' ') switch (m_char)	/* dispatch on control chars */
	{
	    case CR:
	    case FF:
	    case VT:
		return(m_dot.dot);		/* at end of line, for now */
	    case TAB:				/* move to next tab stop */
		m_count = (m_count | tabmask) + 1;
		if (m_count > x) --m_dot.dot;	break;
	    case ESC:				/* ESC prints as $ - one position */
		m_count++;			break;
	    default:				/* other control chars print as ^X - 2 chars */
		m_count += 2;
		if (m_count > x) --m_dot.dot;	break;
	}
	else ++m_count;				/* printing chars take one space */
    }
    return(m_dot.dot);			    /* character position found */
}


/* routine to move up or down n lines on display */
/* finds character location, then does nL, then moves to x position */

int last_dis_dot;
int last_dis_buff;
int last_dis_x;
int last_dis_wrap;

void display_lines()
{
    int dis_buff = (pbuff == &buffs[0]) ? 0 : 1 ;	/* current buffer */
    struct w_data *pw = dis_buff ? &win_2 : &win_1 ;
    int display_on = (WN_scroll != 0 && (!dis_buff || WN_secondary != 0));

    if (display_on != 0		/* screen must be on */
	&& (pbuff != &buffs[last_dis_buff] || pbuff->dot != last_dis_dot) )
    {
	for (last_dis_wrap = 0; last_dis_wrap <= pw->curs_y; last_dis_wrap++)
	    if (!(wlp[pw->curs_y - last_dis_wrap]->cflag & WF_CONT)) break;
	if (last_dis_wrap > pw->curs_y) last_dis_wrap = 0;
	last_dis_x = pw->curs_x;	/* if dot has changed since last we looked... */
	last_dis_buff = dis_buff;
    }
    pbuff->dot += lines(get_value(1));  /* move that many lines */

    if (display_on != 0)
	last_dis_dot = pbuff->dot = x_to_dot(last_dis_x + last_dis_wrap * (WN_width - 2));
}
#ifdef TECOTOOL
/* routine to convert screen x,y coordinates to a character position */
/* called with coordinates and address of a variable */
/* returns char position as value, buffer selection in variable */
/* returns -1 if window not on */

int xy_to_dot(x, y, p_bufsel)
    int x, y;
    int *p_bufsel;
{
    struct w_data *win_p;	/* pointer to window containing mouse pointer */
    struct w_line *row_p;	/* pointer to a row in display data table */
    struct qp m_dot;		/* character pointer */
    unsigned short m_count;	/* character position */
    unsigned char m_char;	/* character */
    int return_val;

    if (WN_scroll != 0 && y < (WN_height - WN_scroll))		    /* if window is on and mouse is in active area */
    {
	*p_bufsel = (WN_secondary != 0 && y > win_1.end) ? 1 : 0;   /* 1 if in buffer 2 */
	win_p = (*p_bufsel) ? &win_2 : & win_1;			    /* point to corresponding window */

/* point to row containing mouse pointer */

	row_p = wlp[y];
	if (y > win_p->last_y)	    /* if mouse is beyond screen active area */
	{
	    return_val = wlp[win_p->last_y]->end + 1;	/* set last char position on screen + 1 */
	    if (return_val > buffs[*p_bufsel].z) return_val = buffs[*p_bufsel].z;
	}
	else if (x <= row_p->col) return_val = row_p->start;	/* at or before first column */
	else if (x >= row_p->n)
	{
	    return_val = row_p->end;			    /* at or after last column */
	    w_setptr(&buffs[*p_bufsel], return_val, &m_dot);	/* if last char is a LF */
	    if (m_dot.p->ch[m_dot.c] == LF) --return_val;   /* point before it */
	}
	else						    /* somewhere in between */
	{
	    (void) w_setptr(&buffs[*p_bufsel], row_p->start, &m_dot);	    /* set pointer to first char on line */
	    m_count = row_p->col;
	    if (row_p->cflag & WF_CONT) m_count += 1;	    /* contin lines start at col 2 */
	    for ( ; m_count < x; fwdc(&m_dot))		/* for each character on line */
	    {
		m_char = m_dot.p->ch[m_dot.c];		/* character */
		if (m_char < ' ') switch (m_char)	/* dispatch on control chars */
		{
		    case CR:				/* CR with no LF in middle of line */
			m_count = 0;			break;
		    case TAB:				/* move to next tab stop */
			m_count = (m_count | tabmask) + 1;
			if (m_count > x) --m_dot.dot;	break;
		    case ESC:				/* ESC prints as $ - one position */
			m_count++;			break;
		    default:				/* other control chars print as ^X - 2 chars */
			m_count += 2;
			if (m_count > x) --m_dot.dot;	break;
		}
		else ++m_count;				/* printing chars take one space */
	    }
	    return_val = m_dot.dot;			/* character position found */
	}
    }
    else return_val = *p_bufsel = -1;	    /* window off: set dot to -1 */
    return(return_val);
}
#endif
