/* TECO for Ultrix   Copyright 1991 Matt Fichtenbaum				*/
/* This program and its components belong to GenRad Inc, Concord MA 01742	*/
/* They may be copied if this copyright notice is included			*/

/* tecotool.c  SunWindows environment for TECO 12/11/91  08.33			*/
/* This program runs a tty subwindow with TECO in it; it fields keyboard	*/
/* and mouse events and converts them to character strings sent to the tty	*/
/* subwindow.  TECO can then interpret these sequences as it wishes.		*/

/* This program was inspired by "emacstool" supplied by the Free Software	*/
/* Foundation (sort of like "Yes, officer, this bag of money I'm holding	*/
/* was inspired by the bank across the street.") and the author acknowledges	*/
/* his indebtedness to FSF.  This program is thus properly in the public domain.*/

/* The window is, by default, opened with the same size as the window in which
 * "tecotool" is run, with colors that the author finds pleasing.  Both size
 * and colors can be specified explicitly using the standard SunWindows
 * "-W" command line arguments.
 *
 * The following key sequences are sent to TECO:
 *  Top and right function keys:
 *	<ESC> [ nnn z or <ESC> [ {ABCD} z, as in the default SunWindows sequences
 *
 *  Top and right function keys, shifted, change the "z" at the end:
 *	Shift	sets bit 0
 *	Ctrl	clears bit 1
 *	Meta	sets bit 2
 *  Note that the cursor keys and the keypad + - Ins Del Enter keys are mapped
 *  at a higher level within SunWindows and cannot be modified by the shift keys.
 *
 *  Mouse clicks are translated into <ESC> [ 3 yyxx n, where
 *	yy, xx  are hexadecimal y and x character coordinates within the window
 *	n	is a character in the range 0x40-0x7f, whose bits are
 *		0:	shift key is down
 *		1:	ctrl key is down
 *		2:	meta key is down
 *		3:	event is left mouse button
 *		4:	event is middle mouse button
 *		5:	event is right mouse button
 */

#include <suntool/sunview.h>
#include <suntool/tty.h>
#include <suntool/ttysw.h>
#include <stdio.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/types.h>

#define BUFFER_SIZE MAXPATHLEN			/* Size of all the buffers */

static short term_rows = 34, term_cols = 80;	/* default window size */
static char key_string[] = "\033[2nnz";		/* function key char string */
static char mouse_string[] = "\033[3yyxxz";	/* mouse button char string */
static char key_up_string[] = "\033[A";		/* cursor key strings */
static char key_dn_string[] = "\033[B";
static char key_rt_string[] = "\033[C";
static char key_lt_string[] = "\033[D";
static char teco_name[] = "te_tt";		/* default run command */
static char title[] = "TECOtool - ";		/* initial title */
static char buffer[BUFFER_SIZE];		/* send to ttysw_input */
extern char *getenv();
int font_width, font_height;			/* For translating pixels to chars */
struct singlecolor cfore = { 255, 255, 255 };	/* foreground and background colors */
struct singlecolor cback[] =
{
    {	  0,   65,   65 },
    {	  0,   30,  130 },
    {	  0,  100,   30 },
    {	  0,   65,   65 },
    {	100,	0,   30 },
    {	130,	0,    0 },
    {	115,	0,  115 },
    {	105,  105,    0 },
};

Frame frame;					/* Base frame for system */
Tty tty_win;					/* tty subwindow running teco */

Icon frame_icon;
static short default_image[258] =	/* make an icon_image for the default frame_icon */
{
#include "teco.icon"
};
mpr_static(icon_image, 64, 64, 1, default_image);
/* Assign a value to a set of keys */
/*
 * Code up the current situation:
 *
 *  1 = SHIFT;
 *  2 = CONTROL;
 *  4 = META;
 *  8 = MS_LEFT;
 * 16 = MS_MIDDLE;
 * 32 = MS_RIGHT;
 */

int button_value (event)
    Event *event;
{
    int retval = 0;

    if (MS_LEFT	  == (event_id(event))) retval = 8;
    if (MS_MIDDLE == (event_id(event))) retval = 16;
    if (MS_RIGHT  == (event_id(event))) retval = 32;

    if (event_shift_is_down(event)) retval += 1;
    if (event_ctrl_is_down(event)) retval += 2;
    if (event_meta_is_down(event)) retval += 4;
    return(retval);
}


/* convert integer to two hex chars */

static void conv_hex(p, n)
    char *p;			    /* destination buffer */
    int n;			    /* value */
{
    static char digits[] = "0123456789abcdef";

    *p++ = digits[(n &= 0xff) >> 4];
    *p = digits[n & 0x0f];
}
/*
 * Filter function to translate selected input events
 * Mouse button events become ESC [ row col button.
 * Function keys: ESC [ 2nn code, where code is 'z' mod by shift keys.
 */

static Notify_value
input_event_filter_function(window, event, arg, type)
    Window window;
    Event *event;
    Notify_arg arg;
    Notify_event_type type;
{
    int ecode;				    /* temp event code */

    if (event_id(event) == WIN_STOP)	    /* UP L1 is the STOP key */
    {					    /* do nothing for now */
	return(NOTIFY_IGNORED);
    }

/* UP L5 & L7 is Expose & Open, let them pass to sunview */
    else if (event_id(event) == KEY_LEFT(5) || event_id(event) == KEY_LEFT(7))
	if(event_is_up(event))
	    return(notify_next_event_func(window, event, arg, type));
	else return(NOTIFY_IGNORED);

    else if (event_is_key_left(event))	    /* ignore other left keys */
	return(NOTIFY_IGNORED);

    else if (event_is_button(event))	    /* do Mouse Button events */
    {
	if (event_is_up(event))
	    return(NOTIFY_IGNORED);
	conv_hex(&mouse_string[3], event_y(event)/font_height);	    /* put y coord into string */
	conv_hex(&mouse_string[5], event_x(event)/font_width);	    /* ... x coord */
	mouse_string[7] = button_value(event) | '@';		    /* ... mouse button/shift status */
	ttysw_input (tty_win, mouse_string, sizeof(mouse_string)-1);
	return NOTIFY_IGNORED;
    }

/* function key events */
    else if (event_is_key_top(event) || event_is_key_right(event))  /* F1..F15, R1..R15 */
    {
	if (event_is_up(event)) return NOTIFY_IGNORED;		    /* down events only */
	if (event_is_key_top(event)) ecode = event_id(event) - KEY_TOP(1) + 24;	    /* top keys */
	else ecode = event_id(event) - KEY_RIGHT(1) + 8;
	key_string[3] = (ecode/10) + '0';
	key_string[4] = (ecode % 10) + '0';
	key_string[5] = 'z';					    /* end code */
	if (event_shift_is_down(event)) key_string[5] |= 1;	    /* shift, control, meta bits */;
	if (event_ctrl_is_down(event)) key_string[5] &= ~2;
	if (event_meta_is_down(event)) key_string[5] |= 4;

	if (key_string[5] == 'z' && (ecode == 15 || ecode == 17 || ecode == 19 || ecode == 21))
	{							    /* unshifted cursor keys */
	    if (ecode == 15) ttysw_input(tty_win, key_up_string, sizeof(key_up_string)-1);
	    else if (ecode == 17) ttysw_input(tty_win, key_lt_string, sizeof(key_lt_string)-1);
	    else if (ecode == 19) ttysw_input(tty_win, key_rt_string, sizeof(key_rt_string)-1);
	    else ttysw_input(tty_win, key_dn_string, sizeof(key_dn_string)-1);
	}
	else ttysw_input(tty_win, key_string, sizeof(key_string)-1);
	return NOTIFY_IGNORED;
    }

    if ((event_is_ascii(event) || event_is_meta(event)) && event_is_up(event))	    /* ignore up events */
	return NOTIFY_IGNORED;

/* Allow arbitrary mapping of ASCII keys */
    if (event_is_ascii(event))
	event_set_id(event, event_id(event));
    return notify_next_event_func (window, event, arg, type);
}
main(argc, argv)
    int argc;
    char **argv;
{
    int error_code;	    /* Error codes */
    char *termcap;	    /* Current TERMCAP value */
    int i;		    /* temp */
    int rc_found = 0;	    /* indicator for "-rc" arg */
    static void get_term_par();
    int pid;		    /* result from fork */
    char *p = argv[0];	    /* pointer to look at invocation name */

    while (*p != '\0') ++p;		    /* find last component of path name */
    while (p > argv[0] && *--p != '/');	    /* find beginning of that component */
    if (*p == '/') ++p;
    if (*p == 't' && *++p == 't' && *++p == 'b' && *++p == '\0')
    {
	if ((pid = fork()) < 0) /* run the window owner in the background */
	{
	    fprintf(stderr, "tt: can't spawn child process\n");
	    exit(1);
	}
	else if (pid != 0) exit(0);	/* parent: exit */
    }

/* this is not background mode, or it is the child - continue */

    get_term_par();		    /* find window's size */
    if (putenv("TERM=sun") != 0)    /* TTY_WIN will be a TERM=sun window */
    {
	fprintf (stderr, "%s: Could not set TERM=sun, using `%s'\n",
	     argv[0], getenv("TERM"));
    }

/*
 * If TERMCAP starts with a slash, it is the pathname of the
 * termcap file, not an entry extracted from it, so KEEP it!
 * Otherwise, it may not relate to the new TERM, so Nuke-It.
 * If there is no TERMCAP environment variable, don't make one.
 */

    termcap = getenv("TERMCAP");
    if (termcap && (*termcap != '/'))
    {
	if (putenv("TERMCAP=") != 0)
	{
	    fprintf (stderr, "%s: Could not clear TERMCAP\n", argv[0]);
	}
    }

/* find command to run as subprocess in window */
    if (!(argv[0] = getenv("TECOTOOL")))    /*  Set teco command name */
	argv[0] = teco_name;

/* Tecotool recognizes -rc <file> as the file to run in the tty window */
    for (argc = 1; argv[argc]; argc++)		    /* Use last one on line */
    {
	if( !(strcmp ("-rc", argv[argc])))	    /* Override if -rc given */
	{
	    rc_found = 1;	    /* set indicator */
	    i = argc;
	    argv[argc--] = 0;	    /* kill the -rc argument */
	    if (argv[i+1])	    /* move to argv[0] and squeeze the rest */
	    {
		argv[0] = argv[i+1];
		for (; argv[i+2]; (argv[i] = argv[i+2], argv[++i]=0) );
	    }
	 }
    }

    strcpy (buffer, title);
    if (rc_found) strncat (buffer, argv[0],	    /* append run command name, if spec'd */
	(BUFFER_SIZE - (strlen (buffer)) - (strlen (argv[0]))) - 1);

    error_code = interpose_on_window(argc,argv);
    if (error_code != 0)	    /* Barf */
    {
	fprintf (stderr, "notify_interpose_event_func returns %d.\n", error_code);
	exit (1);
    }

    window_main_loop (frame);
    return(0);
}
int interpose_on_window(argc, argv)
    int argc;
    char **argv;
{
/* do this first, so arglist can override it */
    frame_icon = icon_create(	ICON_IMAGE,		&icon_image, 0);

/* Build a frame to run in */
    frame = window_create( (Window)NULL, FRAME,
				FRAME_NO_CONFIRM,	TRUE,
				WIN_ROWS,		term_rows,
				WIN_COLUMNS,		term_cols,
				FRAME_FOREGROUND_COLOR, &cfore,
				FRAME_BACKGROUND_COLOR, &cback[getpid() & 7],
				FRAME_INHERIT_COLORS,	TRUE,
				FRAME_LABEL,		buffer,
				FRAME_ICON,		frame_icon,
				FRAME_EMBOLDEN_LABEL,	TRUE,
				FRAME_ARGC_PTR_ARGV,	&argc, argv,
				0);

/* Create a tty with teco in it */
    tty_win = window_create(frame, TTY,
				TTY_QUIT_ON_CHILD_DEATH,    TRUE,
				TTY_BOLDSTYLE,		    TTYSW_BOLD_OFFSET_X,
				TTY_ARGV,		    argv,
				0);

/* ttysw uses pf_default, one must set WIN_FONT explicitly */

    window_set(tty_win, WIN_FONT, pf_default(), 0);
    font_height = (int)window_get (tty_win, WIN_ROW_HEIGHT);
    font_width  = (int)window_get (tty_win, WIN_COLUMN_WIDTH);

    window_set(tty_win,
		 WIN_CONSUME_PICK_EVENTS,
		    WIN_STOP,
		    WIN_MOUSE_BUTTONS, WIN_UP_EVENTS,
		    0,
		 WIN_CONSUME_KBD_EVENTS,
		    WIN_STOP,
		    WIN_ASCII_EVENTS,
		    WIN_LEFT_KEYS, WIN_TOP_KEYS, WIN_RIGHT_KEYS,
		    0,
		 0);

/* Interpose my event function */

    return((int) notify_interpose_event_func(tty_win, input_event_filter_function, NOTIFY_SAFE));
}
/* routine to get terminal height and width from termcap */

static void get_term_par()
{
    char *pname;		    /* pointer to terminal name */
    char tbuff[1024];		    /* termcap buffer */

    pname = getenv("TERM");
    if (pname != 0)
    {
	(void) tgetent(tbuff, pname);	/* get entry */
	term_rows = tgetnum("li");	/* get #lines and #columns and set params */
	term_cols = tgetnum("co");
    }
}
